/* 
 *  $Id: Node.Java,v 1.5 1999/10/04 21:18:13 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_ANTS_PKG;

import JANOS_CORE_PKG.Address;
import JANOS_CORE_PKG.AddressFormatException;
import JANOS_CORE_PKG.BufferHandle;
import JANOS_CORE_PKG.NoSuchNodeAddress;
import JANOS_CORE_PKG.Flow;
import JANOS_CORE_PKG.OutChannel;
import JANOS_CORE_PKG.Credentials;
import JANOS_CORE_PKG.Thread;
import JANOS_CORE_PKG.ThreadPool;
import JANOS_CORE_PKG.ThreadPoolSpec;
import JANOS_CORE_PKG.MemSpec;
import JANOS_CORE_PKG.FlowStartSpec;
import JANOS_CORE_PKG.CommSpaceHandle;

import java.io.IOException;

import java.util.Hashtable;
import java.util.Enumeration;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * The per-flow Node interface.
 * <p>
 * An instance of Node is created in the context of each Application
 * and protocol.
 *
 * %AUTHOR_PAT%
 * %AUTHOR_CKH%
 */
public class Node
{
	private final int address;

	private final Hashtable outChannels = new Hashtable(12, 0.5f);

	private final CommSpaceHandle neighbors;
	PACKAGE final UnknownMidPackets unknownMidPackets;

	/**
	 * Maps groupsIDs to CodeGroups.
	 */
	private final Hashtable GIDtoCodeGroup = new Hashtable();

	/**
	 * Maps MIDs to CapsuleClasses.
	 */
	private final Hashtable MIDtoCapsuleClass = new Hashtable();

	/**
	 * Maps name to CapsuleClass.  This table is used to supply
	 * a PID or MID given the name of a CapsuleClass.
	 */
	private final Hashtable nameToCapsuleClass = new Hashtable(20,0.6f);

	/**
	 * XXX this is static.  That is bad, and wrong.  It shouldn't be
	 * static any more as app ports are Flow-private.
	 */
	private static final Hashtable applications = new Hashtable(5, 0.75f);

	/**
	 * Construct a node.  A node is only constructed by the boot
	 * routine in PrimordialNode, AppBooster or ProtocolBooster.
	 */
	PACKAGE Node(Flow flow, CommSpaceHandle neighbors, int address,
		UnknownMidPackets unknownMidPackets)
	{
		this.address = address;
		this.neighbors = neighbors;
		this.unknownMidPackets = unknownMidPackets;

		// Put node where it can be looked up.
		flow.setObject(this);
		getNeighbors();
	}

	public int getAddress()
	{
		return address;
	}

	/**
	 * Return the Node interface associated with the current flow.
	 * <p>
	 * XXX This method is a hack and should go away once we get
	 * free reign to re-write the basic ANTS interfaces.
	 */
	PACKAGE static Node currentNode()
	{
		return (Node)(Flow.currentFlow().getObject());
	}

	/**
	 * Lookup the neighbor table and construct outchans to
	 * each neighbor.  
	 *
	 */
	private void getNeighbors()
	{
		int index = 0;
		int logicalAddress = neighbors.read(index++).getInt();
		while (logicalAddress != 0)
		{
			Address physicalAddress =
				neighbors.read(index++).getAddress();

			addOutChannel(logicalAddress, physicalAddress);

			// Get logical from next pair...
			logicalAddress = neighbors.read(index++).getInt();
		}
	}

	private void addOutChannel(int logicalAddress,
				   Address physicalAddress)
	{
		TRACE(Trace.ANTS_STARTUP, "addOutChannel maps " +Address.toString(logicalAddress)+ " to " +physicalAddress.toString());

		OutChannel newChan;
		try
		{
			newChan = new OutChannel(Flow.currentFlow(),
						 physicalAddress);
		}
		catch (java.io.IOException ioe)
		{
			panic("OutChannel() threw " +ioe);
		}
		outChannels.put(new Integer(logicalAddress),
				newChan);
	}

	/**
	 * Called by applications that want to instantiate a
	 * locally originiating protocol.
	 * <p>
	 * This is a public interface that we really can't change.
	 */
	public void register(Protocol p)
	{
		TRACE(Trace.ANTS_DL, "Installing protocol " +p);
		try
		{
			p.activate(this);
		}
		catch (Exception e)
		{
			FATALEXCEPTION(e, "gak.");
		}
	}

	/**
	 * Deliver the given capsule to the Application on the
	 * given port number.
	 * <p>
	 * XXX access control?
	 */
	public boolean deliverToApp(Capsule cap, int port)
	{
		Application app = (Application) applications.get(new Integer(port));
		app.receive(cap);
		return true;
	}

	public boolean send(Capsule cap)
	{
		int dest;

		dest = cap.getDst();

		return routeForNode(cap, dest);
	}

	public Thread threadStart(java.lang.Runnable r)
	{
		assert(r);
		assert(Flow.currentFlow());
		assert(Flow.currentFlow().threadPool());

		return Flow.currentFlow().threadPool().threadStart(r);
	}

	public boolean routeForNode(Capsule cap, int destination)
	{
		Xdr xdr;

		TRACE(Trace.ANTS_OUTGOING, "Sending capsule " +cap+ " to " +Address.toString(destination));

		assert(destination != 0);

		// Encode the capsule.
		xdr = cap.encodeCapsuleForSend(this);

		// look up nearest neighbor on the way to destination
		int target = PrimordialNode.nextHopToNode(destination);
		
		if (target == 0) // XXX what to do?
		{
			TRACE(Trace.ANTS_OUTGOING, "  Capsule send failed.  No hop to " +Address.toString(destination));
			return false;
		}
		
		TRACE(Trace.ANTS_OUTGOING, "  Capsule being forwarded to " +Address.toString(target));
		

		// push out the appropriate out-channel
		Object oChanObj = outChannels.get(new Integer(target));
		assert(oChanObj != null);

		// XXX fix when XDR is fixed!
		BufferHandle bh = new BufferHandle(Flow.currentFlow());
		bh.newBuffer(xdr.length());
		bh.setBytes(0, xdr.getBytes(), 0, xdr.length());

		try
		{
			TRACE(Trace.ANTS_OUTGOING, "  Pushing buf " +Debug.dbgPacketSummary(bh, 128));
			((OutChannel)oChanObj).send(bh);
		}
		catch (IOException ioe)
		{
			FATALEXCEPTION(ioe, "when sending on outchannel");
		}

		return true;
	}

	PACKAGE void publishCapsuleClass(CapsuleClass cc)
	{
		Object r;
		assert(cc);
		assert(cc.name());

		TRACE(Trace.ANTS_DL, "publishing CapsuleClass " +cc);

		// Publish in the name -> cc map
		r = nameToCapsuleClass.put(cc.name(), cc);
		if (r != null)
			panic("Re-registered CapsuleClass " +cc.name()+ "!");

		// Publish in the mid -> cc map
		r = MIDtoCapsuleClass.put(cc.methodID, cc);
		assert(r == null);
	}

	PACKAGE void publishCodeGroup(CodeGroup group)
	{
		TRACE(Trace.ANTS_DL, "publishing CodeGroup " + group);
		GIDtoCodeGroup.put(group.groupID, group);
	}

	PACKAGE TypeID findMID(String name)
	{
		CapsuleClass cc;
		cc = (CapsuleClass)nameToCapsuleClass.get(name);
		if (cc == null)
			panic("No CapsuleClass registered under " +name+ " (in findMID()).");

		assert(cc.methodID);
		return cc.methodID;
	}

	PACKAGE TypeID findPID(String name)
	{
		CapsuleClass cc;
		cc = (CapsuleClass)nameToCapsuleClass.get(name);
		if (cc == null)
			panic("No CapsuleClass registered under " +name+ " (in findPID()).");

		assert(cc.protocolID());
		return cc.protocolID();
	}

	PACKAGE CapsuleClass lookupCapsuleClass(TypeID mid)
	{
		TRACE(Trace.ANTS_DL, "looking up " + MIDtoCapsuleClass.get(mid));
		return (CapsuleClass) MIDtoCapsuleClass.get(mid);
	}

	PACKAGE CodeGroup lookupCodeGroup(TypeID gid)
	{
		TRACE(Trace.ANTS_DL, "looking up " + GIDtoCodeGroup.get(gid));
		return (CodeGroup) GIDtoCodeGroup.get(gid);
	}

	PACKAGE static void attachApplication(Application app, int port)
		throws AttachException
	{
		Integer key = new Integer(port);
		
		if (applications.containsKey(key))
			throw new AttachException("Port " +port+ " is busy.");
		
		applications.put(key, app);
	}

	public void sleep(int millis)
		throws InterruptedException
	{
		JANOS_CORE_PKG.Thread.sleep(millis, 0);
	}

	public long time()
	{
		return JANOS_CORE_PKG.Node.time();
	}

	public void log(int level, String msg)
	{
		System.out.println(msg);
	}
}
