/* 
 *  $Id: PrimordialNode.Java,v 1.4 1999/09/07 21:15:56 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_ANTS_PKG;

import JANOS_ANTS_UTILS_PKG.KeyArgs;
import JANOS_ANTS_UTILS_PKG.InvalidKeyArgsException;

import JANOS_ROUTE_PKG.RouteTable;
import JANOS_ROUTE_PKG.RouteEntry;

import JANOS_CORE_PKG.AddressFormatException;
import JANOS_CORE_PKG.Credentials;
import JANOS_CORE_PKG.ClassSpec;
import JANOS_CORE_PKG.CommSpaceHandle;
import JANOS_CORE_PKG.CommSpaceElement;
import JANOS_CORE_PKG.CommQueue;
import JANOS_CORE_PKG.DeliverFunc;
import JANOS_CORE_PKG.DemultiplexKey;
import JANOS_CORE_PKG.Flow;
import JANOS_CORE_PKG.FlowStartSpec;
import JANOS_CORE_PKG.InChannel;
import JANOS_CORE_PKG.MemSpec;
import JANOS_CORE_PKG.NoSuchNodeAddress;
import JANOS_CORE_PKG.OutChannel;
import JANOS_CORE_PKG.ThreadPool;
import JANOS_CORE_PKG.ThreadPoolSpec;

import java.io.IOException;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * The root Node object that boots the system and tracks system-wide
 * state.  This is as opposed to the per-flow Node interface
 * that every protocol/application sees.
 * <p>
 * XXX move the ConfigurationManager stuff into here.
 *
 * %AUTHOR_PAT%
 * %AUTHOR_CKH%
 */
PACKAGE class PrimordialNode
{
	/**
	 * Arguments the PrimordialNode understands and their defaults.
	 */
	private final static String[] defaults = { "-updateRoutes", "false",
						   "-log", "0",
						   "-routes"      , null};

	private static String routefile = "ping.routes"; // XXX XXX XXX
	private static boolean updateRoutes = false;
	private static RouteTable routeTable;
	private static CommSpaceHandle neighborsSpace;

	/**
	 * 32-bit logical address of this Node.
	 * <p>
	 * XXX replace with a real Address object?
	 */
	private static int address;

	// Maps process IDs to CommQueues:
	private static final Hashtable pidToFlowQueue = new Hashtable();

	/**
	 * Global cache of packets for MIDs that haven't been loaded/resolved.
	 *
	 * The Dead Letter Office.
	 */
	PACKAGE static final UnknownMidPackets unknownMidPackets =
		new UnknownMidPackets();

	/**
	 * Start the node on its way.
	 */
	PACKAGE static void start(Vector appConfigs)
		throws IOException, AddressFormatException, NoSuchNodeAddress,
		InterruptedException
	{
		/*
		 * Setup the route table.  Initialize from the routefile
		 * argument.
		 */
		routeTable = new RouteTable(getAddress(), updateRoutes);
		if (routefile != null)
		{
			TRACE(Trace.ANTS_STARTUP, "Reading route table " +routefile);
			routeTable.read(routefile);
			// above throws IOException, AddressFormatException and NoSuchNodeAddress

			routefile = null; // drop the route file name
		}

		/*
		 * Get the list of neighbors into a form the child app
		 * can use.
		 */
		neighborsSpace = fillNeighborSpace();
		
		/*
		 * Create a Node interface for the builtin protocols.
		 */
		Node node = new Node(
			Flow.currentFlow(),
			neighborsSpace,
			address,
			unknownMidPackets);

		assert(address != 0);

		try
		{
			/*
			 * Create the DLProtocol.  
			 *
			 * Manually associate it with a key that matches
			 * just the ANTS header.  This way it will get all
			 * packets that aren't matched by more specific
			 * filters.
			 *
			 * XXX runs with the same ThreadPool as other root flow tasks...
			 * ... might be okay actually (nothing else for those threads).
			 */
			DLProtocol dlp = new DLProtocol();
			
			TRACE(Trace.ANTS_STARTUP, "Installing protocol " +dlp);

			// XXX that's really the UnknownMIDPacketHandler:
			UnknownPacketHandler uph = new UnknownPacketHandler(node);
			DemultiplexKey upKey = new ANEPDemultiplexKey(0, Capsule.ANTSTYPE);
			InChannel ichan = new InChannel(
					Flow.currentFlow(),
					upKey,
					Flow.currentFlow().threadPool(),
					uph);

			/*
			 * Activate the MIDIDRequest/Response parts of
			 * the DLProtocol.  (The other parts
			 * aren't activated because they're inert --
			 * see BuiltinProtocol.)
			 */
			dlp.activate(node);


			/*
			 * Setup the DataProtocol.  Self-forwarding data
			 * packet service is provided by the Node.
			 *
			 * XXX this should be tuned & optimized.
			 */
			DataProtocol dp = new DataProtocol();
			TRACE(Trace.ANTS_STARTUP, "Installing protocol " +dp);
			dp.activate(node);

			// XXX unknown ANEP packet forwarder--- a cut-through channel!.
		}
		catch (Exception e)
		{
			FATALEXCEPTION(e, "Builtin protocol setup failure.");
		}
		
		/*
		 * Start up the applications on this node.
		 */
		Enumeration apps = appConfigs.elements();
		while(apps.hasMoreElements())
		{
			AppConfig a = (AppConfig)apps.nextElement();
			fireUpNewAppFlow(a);
		}

		/*
		 * XXX MIT ANTS supports Node extensions and installs
		 * them at this point.  Since they're unused, we ignore
		 * them for now.
		 */

	}

	/**
	 * Setup one outchan direct to each neighbor node.
	 * Enter each into a hashtable indexed by the logical
	 * address of the neighbor.
	 *
	 * XXX record the logical address of each neighbor in the
	 * neigborsSpace commSpace.  Child protocols and apps
	 * will create a set of OutChannels from this info.
	 * Seems rather weak and ugly, though.  Perhaps push-able
	 * OutChannelHandles would be a good idea?
	 *
	 * XXX this is now redundant with the per-node outchannels,
	 * no?
	 *
	 * XXX Its really, really sad that I call this once for each
	 * Flow I create.  Kinda stupid, no?
	 */
	PACKAGE static CommSpaceHandle fillNeighborSpace()
		throws IOException
	{
		Enumeration neighbors = routeTable.getNeighbors();

		TRACE(Trace.ANTS_ROUTING|Trace.ANTS_STARTUP, "I have " +routeTable.neighborCount()+ " neighbor(s).");

		CommSpaceHandle space = CommSpaceHandle.newCommSpace(
			routeTable.neighborCount() * 2 + 1);

		int idx = 0;
		while (neighbors.hasMoreElements())
		{
			RouteEntry r = (RouteEntry)neighbors.nextElement();

			// Neighbor table should only have immediate neighbors
			assert(r.getLogicalDest() == r.getLogicalNext());
			
			TRACE(Trace.ANTS_ROUTING, "Have immediate neighbor " +r);

			space.write(idx++,
				new CommSpaceElement(r.getLogicalDest()));
			space.write(idx++,
				new CommSpaceElement(r.getPhysicalNext()));
		}
		space.write(idx++, new CommSpaceElement(0));
		
		// XXX ... its bad if this node has no immediate neighbors, m'Kay.
		if (idx == 1)
		{
			panic("This node has no immediate neighbors.  No where to route to!");
		}

		return space;
	}


	/**
	 * Return the logical ANTS address of the next node
	 * on the path to the given addr.
	 */
	PACKAGE static int nextHopToNode(int addr)
	{
		// assert(addr != ..);
		RouteEntry re = routeTable.get(addr);
		if (re == null)
			return 0;

		return re.getLogicalNext();
	}


	/*
	 * Parse a set of KeyArgs arguments.  Adds '-routes
	 * <filename>', '-log <int>' and '-updateRoutes <boolean>'
	 * arguments.  Will throw an InvalidKeyArgsException if any
	 * arguments are not understood by Node or its superclases.
	 * <p>
	 * Defaults '-updateRoutes' to false.  Defaults route
	 * filename to null.
	 *
	 * @see #defaults
	 */
	PACKAGE static void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(defaults);
		
		for (int i = 0; i < k.length(); i++) 
		{
			if (k.key(i).equals("-routes")) 
			{
				routefile = k.arg(i);
				k.strike(i);
			}
			else if (k.key(i).equals("-updateRoutes"))
			{
				updateRoutes = Boolean.valueOf(k.arg(i)).booleanValue();
				k.strike(i);
			}
			else if (k.key(i).equals("-log")) 
			{
				// XXX ignore -log for now
				k.strike(i);
			}
			
		}
		
		KeyArgs.noMore(k);
	}

	public static int getAddress()
	{
		return address; // XXX
	}

	/**
	 * Set this node's logical address.  Only invoked
	 * once during configuration of the system
	 * after the Node object is created.
	 *
	 * @see ConfigurationManager#parseCommands()
	 */
	PACKAGE static void setAddress(int address)
	{
		PrimordialNode.address = address;
	}	

	/**
	 * Add the given MID to the set understood by this node.
	 * This creates a flow to handle the mid (using the
	 * resource defined by the userInfo field).
	 * <p>
	 * XXX Currently the whole method is synchronized.  This
	 * protects the race-prone lookup/create in the pidToFlowQueue
	 * map.  It doesn't need to protect the fireUpNewProtocolFlow.
	 *
	 * @param pid the PID of the new flow.
	 * @param userInfo the user identity/credentials of the new flow
	 * @param mid the MID that caused the flow to get loaded.
	 * @param prevNode the node that sent the initial MID-labeled packet.
	 */
	PACKAGE static synchronized void addMID(TypeID pid,
						UserID userInfo,
						TypeID mid,
						int prevNode)
	{
		CommQueue queue = (CommQueue) pidToFlowQueue.get(pid);
		if(queue == null)
		{
			queue = new CommQueue();
			pidToFlowQueue.put(pid, queue);
			try
			{
				fireUpNewProtocolFlow(pid, queue);
			}
			catch (Throwable th)
			{
				FATALEXCEPTION(th, "Uh-oh.  fireUpNewFlow() blew up.");
			}
		}

		// XXX why do this if PID already registered?
		queue.enqueue(new CommSpaceElement(mid.unwrap()));
		queue.enqueue(new CommSpaceElement(prevNode));
	}

	private static void fireUpNewAppFlow(AppConfig app)
		throws InterruptedException,
		IOException
	{
		CommSpaceHandle csh = fireUpNewFlow(
			new Credentials("application"), // XXX
			new ThreadPoolSpec(null, null), // XXX
			new MemSpec(100), // XXX
			null,
			null,
			AppBooster.class.getName(),
			app.args);
	}

	private static void fireUpNewProtocolFlow(
			TypeID pid,
			CommQueue queue)
		throws InterruptedException,
		IOException
	{
		fireUpNewFlow(
			new Credentials("dlprotocol:" +pid), // XXX
			new ThreadPoolSpec(null, null), // XXX
			new MemSpec(100), // XXX
			pid,
			queue,
			ProtocolBooster.class.getName(),
			null);
	}

	private static CommSpaceHandle fireUpNewFlow(
			Credentials cred,
			ThreadPoolSpec tps,
			MemSpec ms,
			TypeID pid,
			CommQueue queue,
			String className,
			String[] args)
		throws InterruptedException,
		IOException
	{
		CommSpaceHandle csh = CommSpaceHandle.newCommSpace(5);
		csh.write(0, new CommSpaceElement(getAddress()));
		csh.write(1, new CommSpaceElement(neighborsSpace));
		if(pid != null)
			csh.write(2, new CommSpaceElement(pid.unwrap()));
		if(queue != null)
			csh.write(3, new CommSpaceElement(queue));
		csh.write(4, new CommSpaceElement(unknownMidPackets.spaces));

		FlowStartSpec fss = new FlowStartSpec(
			new ClassSpec[0], // XXX XXX XX
			className,
			args,
			csh);

		Runnable th = new Runnable() {
			public void run() {
				panic("termination handler invoked?!");
			}
		};
		
		Flow.currentFlow().newFlow(cred, tps, ms, fss, th);

		return csh;
	}
}
