/*
 *
 * Copyright (c) 1998,1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTS_UTILS_PKG;

import java.util.Vector;
import java.util.Hashtable;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.IOException;

/**
 * This class implements the ByteCodeReader interface
 * based on a "CLASSPATH" which can be given as a parameter.
 *
 * @author Godmar Back (gback@cs.utah.edu)
 */
public
class ClassPathReader
	implements ByteCodeReader
{
    /**
     * Components of this path
     */
    private Vector pathcomps;

    /**
     * Hashtable of already read classes
     */
    private Hashtable	classcache;


    /**
     * Use the system property 'java.class.path' as the default
     * classpath to search.
     */
    public ClassPathReader() 
    {
	this(System.getProperty("java.class.path"));
    }

    /**
     * Build a class path from the specified path string
     *
     * @param pathstr the CLASSPATH to use for future class searches.
     */
    public ClassPathReader(String pathstr) 
    {
	classcache = new Hashtable();
	pathcomps = new Vector(5,2);

	String s = pathstr;
	int i;
	while ((i = s.indexOf(File.pathSeparatorChar)) != -1) {
	    if (i == 0)
		pathcomps.addElement(".");
	    else
		pathcomps.addElement(s.substring(0, i));
	    s = s.substring(i+1);
	}
	pathcomps.addElement(s);
    }

    /**
     * Resolve a class name of the form foo.bar.baz
     * Searches the classpath given to the constructor.
     *
     * @param classname fully qualified class name
     * @return byte[] array of bytes containing .class file
     */
    public byte[] readByteCode(String classname) throws Exception
    {
	// check in cache first
	byte [] p = (byte [])classcache.get(classname);

	if (p != null)
	    return p;

	BufferedInputStream file;
	String cname = classname.replace('.', '/').concat(".class");

	// iterate through all components and look for .class file
	for (int i = 0; i < pathcomps.size(); i++) {

	    String compname = (String)pathcomps.elementAt(i);
	    try {
		// a zip file ?
		ZipFile  zip   = new ZipFile(compname);
		ZipEntry entry = zip.getEntry(cname);

		// avoid throwing and catching a null pointer exception
		if (entry == null)
		    throw new Exception("not in zip file");

		file = new BufferedInputStream(zip.getInputStream(entry));

	    } catch (Exception ie0) {

		try {
		    String fn  = compname.concat("/").concat(cname);

		    // a regular file
		    file = new BufferedInputStream(new FileInputStream(fn));

		} catch (IOException ie1) {
		    continue;
		} 
	    }

	    // now try to read the file
	    p = new byte[file.available()];
	    if (file.read(p) != p.length)
		throw new Exception("Short read for " + classname);

	    classcache.put(cname, p);
	    return p;
	}

	throw new Exception("Couldn't resolve class `" 
		+ classname + "'\n" 
	    	+ " in CLASSPATH " + this);
    }

    /**
     * convert to printable form
     * 
     * @return String describing this class path
     */
    public String toString() {
	return pathcomps.toString();
    }
}
