/*
 * Debug.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_CORE_PKG;

/* XXX Gak.  A weak hack until we've got some native library support for debugging. */
#define NATIVE_SUPPORT_IS_BROKEN

#ifdef DEBUG
/**
 * The package-internal debugging infrastructure.
 * XXX Currently this is public for my convenience.
 * <p>
 * To avoid dependencies on classes that are not automagically
 * loaded into all spaces, I have to jump through several hoops to
 * make sure that non-critical classes are not used.  (In particular,
 * I have avoided the '"string" + String' syntax to avoid implicitly
 * referencing java/lang/StringBuffer.
 * <p>
 * The majority of routines in this file deal with printing directly
 * to stdout/stderr through VM facilities.  See the corresponding
 * native layer for the details on where that output is directed.
 * <p>
 * There are three basic facilities provided by this class, first
 * the <code>assert()</code> methods which really aren't very useful
 * without a pre-processor pass.
 * Second, the primitive print functions and object print functions,
 * because we all love printf-debugging.
 * Third, the flag-based trace primitive for dynamically adjusting
 * the printing of various parts of the Java system.
 * <p>
 * Technically, this class should be final so that folks don't go
 * around sub-classing to.  Too bad.  This is much easier.  :)
 * 
 * %CORE_DOC_COPYRIGHT%
 * %CORE_DOC_LICENSE%
 * %CORE_VERSIONTAG%
 *
 * %AUTHOR_PAT%
 */
public final class Debug
{
	/**
	 * If an assert fails, this routine is called to print the
	 * file and line number information, along with the string
	 * representation of the failed expression, and a backtrace
	 * to the failure.
	 *
	 * @param str the string representation of the expression
	 * @param file the name of the file in which the assert failed
	 * @param line the line number on which the assert failed in <code>file</code>
	 */
	private static void assertFailure(String str, String file, int line)
	{
			print(file);
			print(":");
			printInt(line);
			print(": Assertion failed: ");
			println(str);

			printStackTrace();

			// println(" ---- STATS: ----");
			// // dumpStats();
			// println(" ---- :STATS ----");

			exit();
	}
	
	/**
	 * Check that the boolean expression is true.  Calls assertFailure()
	 * if not.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure
	 */
	public static void assertCheck(boolean expr, String exprStr, String file, int line) 
	{
		if (!expr) 
		{
			assertFailure(exprStr, file, line);
		}
	}

	/**
	 * Check that the given object reference is non-null. Calls assertFailure() if the reference
	 * is null.
	 *
	 * @param expr the expression to check
	 * @param exprStr the string representation of the expression
	 * @param file the name of the file the assert is in
	 * @param line the line number of the file the assert is in
	 * @see #assertFailure
	 */
	public static void assertCheck(Object ref, String exprStr, String file, int line) 
	{
		if (ref == null) 
		{
			assertFailure(exprStr, file, line);
		}
	}

	/**
	 * Print a trace message.
	 *
	 * @param file the name of the file the TRACE is in
	 * @param line the line number of the file the TRACE is in
	 * @param msg the trace message to print
	 */
	public static void pTrace(String file, int line, String msg) 
	{
		System.out.println(file+ ":" +line+ "  " +msg);
	}

	/**
	 * Assert that the given object is locked (synchronized).
	 *
	 * @param obj the object to check for lock status.
	 */
	public static void checkLocked(Object obj, String exprStr, String file, int line)
	{
		return;
#ifndef NATIVE_SUPPORT_IS_BROKEN
		if (!objLocked(obj))
		{
			assertFailure(exprStr, file, line);
		}
#endif
	}
	
	/** 
	 * Test that the given object is locked.
	 */
	private static native boolean objLocked(Object obj);


	/**
	 * Dump various stats about the state of the system.
	 * This method is primarily called by the VM thread
	 * system when it runs out of useful threads to
	 * allocate (i.e., something is deadlocked.)
	 */
	public static void dumpStats()
	{
		// Thread.dumpThreads();
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints a trailing newline.
	 *
	 * @param str The string to print
	 */
#ifndef NATIVE_SUPPORT_IS_BROKEN
	public static native void println(String str);
#endif
	public static void println(String str)
	{
		java.lang.System.err.println(str);
	}
	

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Does not print a trailing newline.
	 *
	 * @param str The string to print
	 */
#ifndef NATIVE_SUPPORT_IS_BROKEN
	public static native void print(String str);
#endif
	public static void print(String str)
	{
		java.lang.System.err.print(str);
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given integer as a string (in base 10).
	 *
	 * @param i the integer to print
	 */
#ifndef NATIVE_SUPPORT_IS_BROKEN
	public static native void printInt(int i);
#endif
	public static void printInt(int i)
	{
		java.lang.System.err.print(i);
	}

	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given integer as a string (in hex).
	 *
	 * @param i the integer to print
	 */
	public static native void printIntHex(int i);

	/**
	 * A native print routines that doesn't require the basic Java clases.
	 * Prints the IPC stages encoded in the given byte.  Doesn't print a
	 * newline.
	 *
	 * @param ops the IPC stages byte to print.
	 */
	public static native void printIPCOps(byte ops);
	
	/**
	 * A native print routine that doesn't require the basic Java classes.
	 * Prints the given char array as a sequence of characters.
	 *
	 * @param array the char array to print
	 * @param start the first element of the char array to print
	 * @param end the last element of the char array to print
	 */
	public static native void printCharArray(char[] array, int start, int end);

	/**
	 * A direct exit.  Do not pass Go; do not collect two hundred dollars.
	 *
	 * TODO: Rather dangerous to leave around.
	 */ 
#ifndef NATIVE_SUPPORT_IS_BROKEN
	public static native void exit();
#endif
	public static void exit()
	{
		java.lang.System.err.println("Debug.exit() invoked by " + Flow.currentFlow().getName());
		java.lang.System.exit(11);
	}
	

	public static String dbgPacketSummary(BufferHandle bufh)
	{
		OffsetLength ol = new OffsetLength();
		Buffer b = bufh.getBuffer(ol);
		return dbgPacketSummary(b, ol.offset, ol.length);
	}

	public static String dbgPacketSummary(BufferHandle bufh, int maxDump)
	{
		OffsetLength ol = new OffsetLength();
		Buffer b = bufh.getBuffer(ol);
		return dbgPacketSummary(b, ol.offset, ol.length, maxDump);
	}

	public static String dbgPacketSummary(UdpBuffer b)
	{
		return dbgPacketSummary(b, 0, b.length());
	}

	public static String dbgPacketSummary(Buffer b, int len)
	{
		return dbgPacketSummary(b.data, 0, len);
	}

	public static String dbgPacketSummary(Buffer b, int offset, int len)
	{
		return dbgPacketSummary(b.data, offset, len);
	}

	public static String dbgPacketSummary(Buffer b, int offset, int len, int maxDump)
	{
		return dbgPacketSummary(b.data, offset, len, maxDump);
	}

	public static String dbgPacketSummary(byte[] b, int offset, int len)
	{
		return dbgPacketSummary(b, offset, len, 128);
	}

	public static String dbgPacketSummary(byte[] b, int offset, int len, int maxDump)
	{
		StringBuffer sbuf = new StringBuffer(50 + (maxDump*4));
		sbuf.append("Packet: len=" +len+ "; offset=" +offset+ "; data=");
		for(int i = 0; i < len; i++)
		{
			if(i >= maxDump)
			{
				sbuf.append("...");
				break;
			}
			// XXX replace this with simple native code!
			// (I want two hex digits per byte).
			if((i % 16) == 0)
				sbuf.append("  ");
			byte val = b[i];
			byte hi = (byte)((int)(val&0xFF) >> 4);
			sbuf.append(Integer.toHexString(hi&0xF));
			sbuf.append(Integer.toHexString(val&0xF));
			sbuf.append(" ");
		}
		return sbuf.toString();
	}

	/**
	 * A native routine to print pertinent information about the
	 * given thread.
	 *
	 * @param th a Thread object to print information about.
	 */
	public static native void threadInfo(Thread th);

	/**
	 * Print the stack trace to the current point of execution in the current
	 * thread.
	 */
#ifndef NATIVE_SUPPORT_IS_BROKEN
	public native static void printStackTrace();
#endif
	public static void printStackTrace()
	{
		Throwable t = new Throwable();
		t.printStackTrace(java.lang.System.err);
	}
	
	public static InternalError fatalException(Throwable t, String msg)
	{
		t.printStackTrace(java.lang.System.err);
		return dPanic(msg);
	}
	
	/**
	 * Off-the-straight-and-narrow.  Sort of a tacky "panic()".
	 * Prints "Off the straight and narrow!" and kills the VM.
	 *
	 * @see otsan(String)
	 */
	public static void otsan()
	{
		otsan("Off the Straight and Narrow!");
	}

	/**
	 * Off-the-straight-and-narrow.  Prints the provided string and kills the VM.
	 * Also prints a stack trace of the current thread.
	 *
	 * @param str the string to print before dying.
	 */
	public static void otsan(String str)
	{
		print("Off the straight and narrow! (Gak.):");
		println(str);
		printStackTrace();
		exit();
	}

	/**
	 * A generic not-implemented print-and-die function.
	 * Prints a stack trace before dying.
	 */
	public static void notImplemented()
	{
		println("Function not implemented");
		printStackTrace();
		exit();
	}

	/**
	 * A less generic not-implemented print-and-die fucntion.
	 * Prints the provided string and a backtrace before killing the VM.
	 *
	 * @param expl the explaination for this rude behavior.
	 */
	public static InternalError notImplemented(String expl)
	{
		print("NOT IMPLEMENTED: ");
		println(expl);
		printStackTrace();
		exit();
		return new InternalError("exit() returned in Debug.notImplemented()");
	}

	/**
	 * A generic panic() function.  Prints the provided string,
	 * a backtrace, and then kills the VM.
	 * <p>
	 * Whee!  This returns an InternalError so that the caller
	 * might throw the result giving the compiler the impression
	 * that dPanic() will never return.  Hehehe.
	 *
	 * @param str The panic message to use as an excuse for killing the machine.
	 */
	public static InternalError dPanic(String str)
	{
		print("PANIC: ");
		println(str);
		printStackTrace();
		exit();
		return new InternalError("exit() returned in Debug.dPanic");
	}

}
#endif /*DEBUG*/
