/*
 * FlowBase.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_LIB_PKG;

import JANOS_CORE_PKG.Flow;
import JANOS_CORE_PKG.ThreadPool;
import JANOS_CORE_PKG.FlowRun;
import JANOS_CORE_PKG.BufferHandle;
import JANOS_CORE_PKG.CommSpaceHandle;

/**
 * A protocol should override FlowBase so that a flow can
 * be created for it.  When the flow is created, the overridden start()
 * method is called.
 * <p>
 * The threadPool, flow, and args variables are guaranteed to
 * be set before start() is called, but may not be set when
 * the constructor is called.
 */
public abstract class FlowBase extends FlowRun
{
	public Flow flow;
	public ThreadPool threadPool; 
	public String[] args;
	public CommSpaceHandle commSpaceHandle;

	protected FreeList bufferHandleFreeList = new FreeList();

	/**
	 * This sets the threadPool, flow, and args fields and then
	 * calls start().
	 */
	public void run(Flow flow,
			ThreadPool threadPool,
			String[] args,
			CommSpaceHandle commSpaceHandle)
		throws Exception
	{
		this.flow = flow;
		this.threadPool = threadPool;
		this.args = args;
		this.commSpaceHandle = commSpaceHandle;
		start();
	}

	/**
	 * @allocation: Allocates a new BufferHandle() if one
	 * is not available in the Flow's BufferHandleFreeList.
	 */
	public synchronized BufferHandle newBufferHandle()
	{
		BufferHandle b = (BufferHandle) bufferHandleFreeList.getFree();
		if(b == null) b = new BufferHandle(flow);
		return b;
	}

	/**
	 * Adds the BufferHandle to a freeList, so that it can
	 * be reused, and calls reset and removeFromList on
	 * the handle to clear its state.
	 */
	public synchronized void recycleBufferHandle(BufferHandle b)
	{
		b.reset();
		b.removeFromList();
		bufferHandleFreeList.addFree(b);
	}

	public abstract void start() throws Exception;
}

