/*
 * UdpTest.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

package JANOS_TESTS_PKG;

import JANOS_CORE_PKG.FlowRun;
import JANOS_CORE_PKG.Flow;
import JANOS_CORE_PKG.ThreadPool;
import JANOS_CORE_PKG.MemSpec;
import JANOS_CORE_PKG.Node;
import JANOS_CORE_PKG.OutChannel;
import JANOS_CORE_PKG.InChannel;
import JANOS_CORE_PKG.DeliverFunc;
import JANOS_CORE_PKG.DemultiplexKey;
import JANOS_CORE_PKG.BufferHandle;
import JANOS_CORE_PKG.Address;
import JANOS_CORE_PKG.AddressFactory;
import JANOS_CORE_PKG.NoSuchNodeAddress;
import JANOS_LIB_PKG.FlowBase;
import JANOS_LIB_PKG.SimpleDemultiplexKey;
#ifdef BENCH
import JANOS_CORE_PKG.CycleTimer;
import JANOS_CORE_PKG.CycleTimerTable;
#endif
import java.net.InetAddress;
import java.io.IOException;

public class UdpTest
	extends FlowBase
{
	public UdpTest()
	{
	}

#ifndef BENCH
	static final int port1 = 33001;
	static final int port2 = 33002;
#else //BENCH
	static final int port1 = 33001;
	static final int port2 = 33001;
#endif //BENCH
	static final byte[] key1 = {5, 4, 3, 2};
	static final byte[] key2 = {5, 4, 3, 1};
#ifdef BENCH
	static final int count = 7000;
#else //BENCH
	static final int count = 100;
#endif //BENCH

#ifdef BENCH
	public static CycleTimer emptyTimer = new CycleTimer("empty");
	public static CycleTimer totalTimer = new CycleTimer("total");
	public static CycleTimer receive1Timer = new CycleTimer("receive1");
	public static CycleTimerTable timeTable = new CycleTimerTable();
#endif

	public void start()
	{
		try 
		{
#ifdef BENCH
			timeTable.addElement(emptyTimer);
			timeTable.addElement(totalTimer);
			timeTable.addElement(receive1Timer);
			timeTable.addElement(InChannel.processTimer);
			timeTable.addElement(OutChannel.udpSendTimer);
			timeTable.addElement(InChannel.sendToRecvTimer);
			timeTable.addElement(InChannel.recvToSendTimer);
			timeTable.addElement(InChannel.recvToRecvTimer);
			timeTable.addElement(InChannel.threadSchedTimer);
			timeTable.addElement(ThreadPool.threadSchedTimer);
			timeTable.addElement(ThreadPool.threadDoneTimer);
#endif //BENCH
			start1();
		}
		catch(Throwable e)
		{
			e.printStackTrace();
			throw new RuntimeException("test failed");
		}
	}

	public void start1()
		throws
			IOException,
			InterruptedException,
			NoSuchNodeAddress
	{
		AddressFactory af = Address.getAddressFactory();
		String iAddr = InetAddress.getLocalHost().getHostAddress();
		Address addr1 = af.newAddress(1, iAddr, port1);
		Address addr2 = af.newAddress(2, iAddr, port2);

#ifndef BENCH
		startMsgLoop(port1);
		startMsgLoop(port2);
#else //BENCH
		startMsgLoop(port1);
#endif //BENCH

		Thread.sleep(2000, 0); // XXX hack to make sure threads ready

		OutChannel out1 = makeOutChannel(addr2);
		OutChannel out2 = makeOutChannel(addr1);
		InChannel in1 = makeInChannel(port1, out1, key1, key2);
		InChannel in2 = makeInChannel(port2, out2, key2, key1);

		BufferHandle b = new BufferHandle(this.flow);
		b.newBuffer(key2.length + 4);
		b.setBytes(4, key2, 0, key2.length);

		out1.send(b);

		recycleBufferHandle(b);
	}

	private void startMsgLoop(final int port)
		throws IOException
	{
		Runnable r = new Runnable()
		{
			public void run()
			{
				try{InChannel.msgLoop(port);}
				catch(Throwable t) {t.printStackTrace();}
			}
		};

		threadPool.threadStart(r);
	}

	private OutChannel makeOutChannel(Address outAddr)
		throws IOException
	{
		return new OutChannel(
			flow,
			outAddr);
	}

	private InChannel makeInChannel(final int inPort,
					final OutChannel replyChannel,
					final byte[] recvKey,
					final byte[] replyKey)
		throws IOException
	{
		final byte[] data = new byte[replyKey.length + 4];
		System.arraycopy(replyKey, 0, data, 4, replyKey.length);

		DeliverFunc deliver = new DeliverFunc()
		{
			// XXX this only works if packets are
			// reliable and ordered

			public void receive(BufferHandle packet, int _)
			{
				try {receive1(packet);}
				catch(IOException e) {e.printStackTrace();}
			}

			public void receive1(BufferHandle packet)
				throws IOException
			{
#ifdef BENCH
				totalTimer.end();
				totalTimer.start();
				emptyTimer.start();
				emptyTimer.end();
				receive1Timer.start();
#endif //BENCH
				int i = packet.getInt(0);

#ifndef BENCH
				Node.println("port " + inPort
					+ " recv:" + i);
#endif //BENCH

				if(i <= count)
				{
					packet.setInt(0, i+1);
					packet.setBytes(4, replyKey, 0, replyKey.length);
					replyChannel.send(packet);
				}

				if(i >= count)
				{
					sendExitPacket(replyChannel);
				}
#ifdef BENCH
				receive1Timer.end();
				if((i + 1) % 1000 == 0)
				{
					System.out.println();
					timeTable.print(System.out);
					timeTable.resetAll();
				}
#endif //BENCH
			}
		};

		return new InChannel(
			flow,
			new SimpleDemultiplexKey(4, recvKey),
			threadPool,
			deliver);
	}

	private void sendExitPacket(OutChannel c)
		throws IOException
	{
		BufferHandle b = new BufferHandle(this.flow);
		b.newBuffer(InChannel.exitPacket.length);
		b.setBytes(0, InChannel.exitPacket, 0,
			   InChannel.exitPacket.length);

		c.send(b);

		recycleBufferHandle(b);
	}
}
